## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, rietbroek@utwente.nl


# This file contains NEQ attributes of the xinv package and lookup methods to determine normal equation components
from collections import namedtuple
from xinv.core.logging import xinvlogger 

#centrally define the xinv attributes
_xinvtype=namedtuple("_xinvtype","unk_co aux_co grp_id_co grp_seq_co N COV rhs solest stdsolest x0 ltpl sigma0 nobs npara")

xinv_tp=_xinvtype("unk_coord","aux_coord","group_id_coord","group_seq_coord","N","COV", "rhs","solest","solest_std","aprioriVec","ltPl","sigma0","nobs","npara")


_xinvstate=namedtuple("_xinvstate","linked unlinked init apri post symU symL cholU cholL")
xinv_st=_xinvstate("linked","unlinked","initial","apriori","posteriori","SymUpper","SymLower","CholeskyUpper","CholeskyLower")


def xinv_attrs(xitype,xistate,xidescr):
    return {"xinv_type":xitype,"xinv_state":xistate,"xinv_description":xidescr}

def rhs_attrs():
    return xinv_attrs(xinv_tp.rhs,xinv_st.init,"right hand side vector")

def x0_attrs():
    return xinv_attrs(xinv_tp.x0,xinv_st.apri,"Apriori solution estimate")

def solest_attrs():
    return xinv_attrs(xinv_tp.solest,xinv_st.init,"solution estimate")

def stdsolest_attrs():
    return xinv_attrs(xinv_tp.stdsolest,xinv_st.init,"Standard deviations of the solution estimate")


def N_attrs(lower=0):
    if lower == 0:
        state=xinv_st.symU
    else:
        state=xinv_st.symL

    return xinv_attrs(xinv_tp.N,state,"Normal matrix")

def Chol_attrs(lower=0):
    if lower == 0:
        state=xinv_st.cholU
    else:
        state=xinv_st.cholL
    return xinv_attrs(xinv_tp.N,state,"Cholesky decomposition")

def cov_attrs(lower=0):
    if lower == 0:
        state=xinv_st.symU
    else:
        state=xinv_st.symL
    return xinv_attrs(xinv_tp.COV,state,"Error Covariance matrix")

def islower(mat):
    """Determine if the xinv_type_str is a lower triangular matrix"""
    state=mat.attrs['xinv_state']
    if state == xinv_st.symL or state == xinv_st.cholL :
        return 1
    elif state == xinv_st.symU or state == xinv_st.cholU :
        return 0
    else:
        raise ValueError(f"Unknown xinv state: {state}. Must be one of {xinv_st.symL}, {xinv_st.symU}, {xinv_st.cholL}, {xinv_st.cholU}")

def ltpl_attrs(state=xinv_st.apri):
    return xinv_attrs(xinv_tp.ltpl,state,"Least squares cost function values")

def sigma0_attrs(state=xinv_st.apri):
    return xinv_attrs(xinv_tp.sigma0,state,"apriori/posteriori standard deviation error scale")

def nobs_attrs():
    return xinv_attrs(xinv_tp.nobs,xinv_st.init,"number of original observations")

def npara_attrs():
    return xinv_attrs(xinv_tp.npara,xinv_st.init,"number of unknown parameters (explicit and implicit)")

def xunk_coords_attrs(state=xinv_st.unlinked):
    """

    Parameters
    ----------
    state : str, optional
        state of the coordinate (linked or unlinked). 'unlinked' denotes that the coordinate sequence may be out of sync with the actual coordinate
        
    Returns
    -------
    xinv_attrs : dict   
    a dictionary with xinv-compliant attributes
    """
    return xinv_attrs(xinv_tp.unk_co,state,"xinv unknown parameter coordinate")

def aux_coords_attrs(state=xinv_st.unlinked):
    return xinv_attrs(xinv_tp.aux_co,state,"xinv auxiliary right hand side/solest coordinate")

def group_id_attrs(state=xinv_st.unlinked):
    return xinv_attrs(xinv_tp.grp_id_co,state,"xinv unknown parameter coordinate indicating the group ids")

def group_seq_attrs(state=xinv_st.unlinked):
    return xinv_attrs(xinv_tp.grp_seq_co,state,"xinv unknown parameter coordinate indicating the sequence within a group")

def find_component(dsneq,component):
    components=[ky for ky in dsneq.keys()]+[ky for ky in dsneq.coords.keys()]
    try:
        compname=[ky for ky in components if "xinv_type" in dsneq[ky].attrs and dsneq[ky].attrs['xinv_type'] == component ]
        return dsneq[compname[0]]
    except KeyError:
        raise KeyError(f"Cannot find {component} component in NEQ dataset")
    except IndexError:
        raise KeyError(f"Cannot find {component} component in NEQ dataset")

def find_components(dsneq,components):
    out=[]
    for comp in components:
        try:
            comp_=find_component(dsneq,comp)
        except KeyError:
            xinvlogger.warning(f"Cannot find {comp} component in NEQ dataset, setting to None")
            comp_=None

        out.append(comp_)
    return out


def find_neq_components(dsneq):
    components=[xinv_tp.N,xinv_tp.rhs,xinv_tp.x0,xinv_tp.ltpl,xinv_tp.sigma0,xinv_tp.nobs,xinv_tp.npara]
    return find_components(dsneq,components)

def find_sol_components(dssol):
    components=[xinv_tp.COV,xinv_tp.solest,xinv_tp.ltpl,xinv_tp.sigma0,xinv_tp.nobs,xinv_tp.npara]
    return find_components(dssol,components)



def find_xinv_coords(dsneq,exclude=None,include=None,state=None):
    if include is not None and exclude is not None:
        raise ValueError("Cannot specify both include and exclude in the same call")

    xinvcoords={}
    coordtypes=[xinv_tp.unk_co,xinv_tp.aux_co,xinv_tp.grp_id_co,xinv_tp.grp_seq_co]
    if exclude is not None:
        #make a subselection
        coordtypes=[coordtype for coordtype in coordtypes if coordtype not in exclude]
    elif include is not None:
        #replace with includes only 
        coordtypes=include

    for coordname in dsneq.coords:
        if "xinv_type" in dsneq[coordname].attrs and dsneq[coordname].attrs['xinv_type'] in coordtypes:
            if state is not None and dsneq[coordname].attrs['xinv_state'] != state:
                continue
            xinvcoords[coordname]=dsneq[coordname]

    return xinvcoords

def find_xinv_unk_coord(dsneq):

    xunk_co=find_xinv_coords(dsneq,include=[xinv_tp.unk_co],state=xinv_st.linked)
    if len(xunk_co)> 1:
        raise ValueError("Ambiguous linked unknown coordinate found")
    return next(iter(xunk_co.values()))

def find_xinv_group_coords(dsneq):
    """
    Find the linked group id and sequence coordinates in a xinv dataset

    Parameters
    ----------
    dsneq : xarray.Dataset
        The dataset to search in
    Returns
    -------
    group_id_co,group_seq_co : xarray.DataArray
        The group id and sequence coordinates, or None if not found
    """

    grp_co=find_xinv_coords(dsneq,include=[xinv_tp.grp_id_co,xinv_tp.grp_seq_co],state=xinv_st.linked)
    grpid_co=None
    grpseq_co=None
    for ky,co in grp_co.items():
        if co.attrs['xinv_type'] == xinv_tp.grp_id_co:
            grpid_co=co
        elif co.attrs['xinv_type'] == xinv_tp.grp_seq_co:
            grpseq_co=co

    return grpid_co,grpseq_co


def get_xunk_size_coname(dsneq):
    """Retrieve the size and name of the currently linked unknown coordinate
    """
    xunk_co=find_xinv_coords(dsneq,include=[xinv_tp.unk_co],state=xinv_st.linked)
    xunkconame=next(iter(xunk_co))
    return dsneq.sizes[xunk_co[xunkconame].dims[0]],xunkconame


def unlink(davar):
    """
    Unlink a variable by changing its state to unlinked
    Parameters
    ----------
    davar : xarray.DataArray
        The data array to unlink.

    Returns
    -------
    None.
    """
    change_state(davar,xinv_st.unlinked)

def link(davar):
    """
    Link a variable by changing its state to unlinked
    Parameters
    ----------
    davar : xarray.DataArray
        The data array to unlink.

    Returns
    -------
    None.
    """
    change_state(davar,xinv_st.linked)

def change_state(davar,state):
    """
    Change the state of a variable to a new state
    Parameters
    ----------
    davar : xarray.DataArray
        The data array to change the state of.
    state : str
        The new state to set.

    Returns
    -------
    None.
    """
    if "xinv_state" in davar.attrs:
        davar.attrs["xinv_state"]=state


def get_state(davar,raiseError=True):
    """
    Get the state of a coordinate variable
    Parameters
    ----------
    davar : xarray.DataArray
        The data array to get the state of.

    Returns
    -------
    xinv_st
        The state of the data array.
    """
    if "xinv_state" in davar.attrs:
        return davar.attrs["xinv_state"]
    else:
        if raiseError:
            raise ValueError("No xinv_state attribute found")
        else:
            return None

def get_type(davar,raiseError=True):
    """
    Get the type of a coordinate variable
    Parameters
    ----------
    davar : xarray.DataArray
        The data array to get the state of.

    Returns
    -------
    xinv_st
        The state of the data array.
    """
    if "xinv_type" in davar.attrs:
        return davar.attrs["xinv_type"]
    else:
        if raiseError:
            raise ValueError("No xinv_type attribute found")
        else:
            return None



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2022 Roelof Rietbroek, rietbroek@utwente.nl

import xarray as xr

from xinv.neq import solve as neqsolve
from xinv.neq import transform as neqtransform
from xinv.neq import groupreduce,reduce,ireduce

from xinv.neq import fix,ifix,groupfix
from xinv.neq import set_x0 as neqset_x0
from xinv.neq import neqadd
from xinv.neq import zeros as neqzeros
from xinv.neq.build import build_normal as neqbuild_normal

from xinv.core.grouping import get_group,reindex_groups,rename_groups

from xinv.core.attrs import find_xinv_coords,xinv_tp,xinv_st

@xr.register_dataarray_accessor("xi")
class InverseDaAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
    
    def build_normal(self,fwdop,ecov=1,**kwargs):
        """Builds a normal equation system from forward operators, data and an accompanying covariance"""
        
        #create the normal equation system
        return neqbuild_normal(fwdop,self._obj,ecov=ecov,**kwargs)

@xr.register_dataset_accessor("xi")
class InverseDsAccessor:
    def __init__(self, xarray_obj):
        self._obj = xarray_obj
    
    def transform(self,fwdop,**kwargs):
        return neqtransform(self._obj,fwdop,**kwargs) #transform the normal equation system using a forward operator
    
    def solve(self,inplace=False):
        return neqsolve(self._obj,inplace) #solve the normal equation system
    
    def reduce(self,keep=False,**kwargs):
        return reduce(self._obj,keep=keep,**kwargs) #reduce  parameters from the normal equation system
    def ireduce(self,idx,keep=False):
        return ireduce(self._obj,idx=idx,keep=keep) #reduce  parameters by index
    
    def groupreduce(self,groupname,keep=False):
        return groupreduce(self._obj,groupname,keep=keep) #reduce  parameters by groupname index
    
    def fix(self,keep=False,**kwargs):
        return fix(self._obj,keep=False,**kwargs) #remove parameters from the normal equation system (fix them to their current apriori values)
    
    def ifix(self,idx,keep=False):
        """fix by index"""
        return ifix(self._obj,idx=idx,keep=keep) #remove parameters from the normal equation system (fix them to their current apriori values)
    
    def groupfix(self,groupname,keep=False):
        return groupfix(self._obj,groupname,keep=keep)

    def set_x0(self,dax0,is_delta=False,inplace=False):
        return neqset_x0(self._obj,dax0,is_delta,inplace) #change apriori values

    def add(self,dsneqother):
        return neqadd(self._obj,dsneqother) #add/merge another normal equation system

    def get_group(self,group_name):
        return get_group(self._obj,group_name)

    def reindex_groups(self,group_dim=None,assoc_coords=None):
        """
        Reindex/rebuild the group coordinates and multinded in the dataset to match the original coordinates
        """
        return reindex_groups(self._obj,group_dim,assoc_coords)
    def rename_groups(self,grpmap):
        return rename_groups(self._obj,grpmap)

    @staticmethod
    def neqzeros(rhsdims,coords,lower=0):
        return neqzeros(rhsdims=rhsdims,coords=coords,lower=lower)

    def unknown_dim(self):
        
        """
        Convenience function to retrieve the name of the currently linked unknown coordinate dimension
        
        Returns 
        -------
        str
            The name of the currently linked unknown coordinate dimension
        
        """
        xunk_co=find_xinv_coords(self._obj,include=[xinv_tp.unk_co],state=xinv_st.linked)
        if len(xunk_co)!=1:
            raise ValueError("No or ambiguous linked unknown coordinate found")
        return next(iter(xunk_co.values())).dims[0]

    def unknown_size(self):
        """
        Convenience function to retrieve the size of the currently linked unknown coordinate
        
        Returns 
        -------
        int
            The size of the currently linked unknown coordinate
        
        """
        xunk_co=find_xinv_coords(self._obj,include=[xinv_tp.unk_co],state=xinv_st.linked)
        if len(xunk_co)!=1:
            raise ValueError("No or ambiguous linked unknown coordinate found")
        return self._obj.sizes[next(iter(xunk_co.values())).dims[0]]



## Permissions: See the  xinv  license file https://raw.githubusercontent.com/strawpants/xinv/master/LICENSE
## Copyright (c) 2025 Roelof Rietbroek, r.rietbroek@utwente.nl

import xarray as xr
import numpy as np
from xinv.core.attrs import find_xinv_coords,find_neq_components, xunk_coords_attrs,xinv_tp,xinv_st

from xinv.core.logging import xinvlogger

from xinv.core.tools import find_ilocs
from xinv.core.grouping import find_group_coords,build_group_coord

def neqadd(dsneq:xr.Dataset, dsneqother:xr.Dataset):
    """ merge two normal equation systems"""
    try:
        N1,rhs1,x01,ltpl1,sigma01,nobs1,npara1=find_neq_components(dsneq)
        N2,rhs2,x02,ltpl2,sigma02,nobs2,npara2=find_neq_components(dsneqother)
    except KeyError:
        raise RuntimeError("Cannot find all NEQ components, are the appropriate xinv_ attributes set?")
    
    if N1.name != N2.name or rhs1.name != rhs2.name or ltpl1.name != ltpl2.name or sigma01.name != sigma02.name or nobs1.name != nobs2.name or npara1.name != npara2.name:
        xinvlogger.warning("Normal system variables have inconsistent variable names, using names from the first system")

    #check for uniform sigma scaling and raise an error if not
    sigma_ratio=sigma01/sigma02
    if len(sigma_ratio.shape) == 0:
        sigma_ratio_single=sigma_ratio.item()
        hasauxdim=False 
    else:
        sigma_ratio_single=sigma_ratio[0].item()
        if not (sigma_ratio == sigma_ratio_single).all():
            raise ValueError(f"Merging NEQs with a common normal matrix but with varying sigma0 ratios over the auxiliary dimensions is not possible")
        hasauxdim=True


    var2_ratio=sigma_ratio_single**2
    #check matrix compatibility
    #check if N is in upper triangular or lower triangular form
    if N1.attrs['xinv_state']!=N2.attrs['xinv_state']:
        raise RuntimeError(f"Can't currently merge NEQs with different xinv_state:{N1.attrs['xinv_state']} vs {N2.attrs['xinv_state']}")
    
    unkdim1=N1.dims[0]
    unkdim2=N2.dims[0]
    
    #find out the unique unknown parameters
    # unique_unk_coord=pd.MultiIndex.from_tuples(np.unique(np.concatenate([N1[unkdim1].data,N2[unkdim2]])))
    unique_unk_coord=xr.DataArray(np.unique(np.concatenate([N1[unkdim1].data,N2[unkdim2]])),dims=[unkdim1],name=unkdim1)
    unique_unk_coord.attrs.update(xunk_coords_attrs(state=xinv_st.linked))

    #find the auxiliary dimensions (ignore the unknown parameter dimension, and group_id/and seq)
    
    xinvcoords=find_xinv_coords(dsneq,exclude=[xinv_tp.grp_id_co,xinv_tp.grp_seq_co])
    group_id_co,group_seq_co,_=find_group_coords(dsneq)

    #check if the group id and sequence coordinates are present in the first system
    if group_id_co is not None and group_seq_co is not None:
        #turn the unknown coordinate into a multiindex
        grp_co=build_group_coord(unique_unk_coord,dim=unkdim1,group_id_name=group_id_co.name,group_seq_name=group_seq_co.name)
        xinvcoords[unkdim1]=grp_co[unkdim1]
        xinvcoords[group_id_co.name]=grp_co[group_id_co.name]
        xinvcoords[group_seq_co.name]=grp_co[group_seq_co.name]
    else:

        #replace the unknow coordinate with the union version
        xinvcoords[unkdim1]=unique_unk_coord
    #add the proper attributes

    
    
    
    xinvcoordsother=find_xinv_coords(dsneqother,exclude=[xinv_tp.grp_id_co,xinv_tp.grp_seq_co])
    #possibly update with complementary coordinates from the second system
    xinvcoords.update({ky:coord for ky,coord in xinvcoordsother.items() if ky not in xinvcoords.keys()})
    #allocate space for the combined output normal equation system and use system one as the base template
    dsneq_merged=xr.Dataset.xi.neqzeros(rhsdims=rhs1.dims,coords=xinvcoords)
    #for some reason the multindex coordinate attributes do not get properly propagated
    #so make sure they are added
    for key,coord in xinvcoords.items():
        if not dsneq_merged[key].attrs:
            dsneq_merged[key].attrs.update(coord.attrs)

    #rename the variables to the same name as in the first system
    renamedict=dict(N=N1.name,rhs=rhs1.name,ltpl=ltpl1.name,sigma0=sigma01.name,nobs=nobs1.name,npara=npara1.name)
    dsneq_merged=dsneq_merged.rename(renamedict)
    
    #get links to the combined normal equation system
    Nout,rhsout,x0out,ltplout,sigma0out,nobsout,nparaout=find_neq_components(dsneq_merged)
    #Note: we need to be very careful when assigning values to the above views so that we insert data at the existing memory locations (!!), rather than replace the views alltogether
    #add stuff from the first normal equation system
    
    idx1=find_ilocs(dsneq_merged,unkdim1,N1[unkdim1].data)
    Nout[idx1,idx1]=N1.data
    rhsout[{unkdim1:idx1}]=rhs1.data
    
    #note use [()] to replace the value in place
    ltplout[()]=ltpl1
    sigma0out[()]=sigma01
    nobsout[()]=nobs1
    nparaout[()]=np.int64(dsneq_merged.sizes[unkdim1])
    #possibly add reduced parameters from the first system
    nreduced=npara1-dsneq.sizes[unkdim1]
    nparaout+=nreduced.astype(np.int64)


    
    #add stuff from the second normal equation system
    idx2=find_ilocs(dsneq_merged,unkdim2,N2[unkdim2].data)

    
    
    #Note that we need to take care of possible different sigma0's
    if var2_ratio != 1.0:
        xinvlogger.info(f"Applying a rescaling of {var2_ratio} to the second normal equation system to compensate for the different apriori variances")
    
    Nout[idx2,idx2]+=var2_ratio*N2.data
    
    rhsout[{unkdim1:idx2}]+=var2_ratio*rhs2.data
    ltplout+=var2_ratio*ltpl2
    #note dsneq_merged['sigma0'] stays the same 
    nobsout+=nobs2
    #possibly add the amount of reduced parameters from the second system
    nreduced=npara2-dsneqother.sizes[unkdim2]
    nparaout+=nreduced.astype(np.int64)

    
    return dsneq_merged                     
    


# def add(dsneq:xr.Dataset, dsneqother:xr.Dataset):
    # """ merge two normal equation systems"""
    # try:
        # N1,rhs1,x01,ltpl1,sigma01,nobs1,npara1=find_neq_components(dsneq)
        # N2,rhs2,x02,ltpl2,sigma02,nobs2,npara2=find_neq_components(dsneqother)
    # except KeyError:
        # raise RuntimeError("Cannot find all NEQ components, are the appropriate xinv_ attributes set?")
    
    # if N1.name != N2.name or rhs1.name != rhs2.name or ltpl1.name != ltpl2.name or sigma01.name != sigma02.name or nobs1.name != nobs2.name or npara1.name != npara2.name:
        # xinvlogger.warning("Normal system variables have inconsistent variable names, using names from the first system")

    # #check for uniform sigma scaling and raise an error if not
    # sigma_ratio=sigma01/sigma02
    # if len(sigma_ratio.shape) == 0:
        # sigma_ratio_single=sigma_ratio.item()
        # hasauxdim=False 
    # else:
        # sigma_ratio_single=sigma_ratio[0].item()
        # if not (sigma_ratio == sigma_ratio_single).all():
            # raise ValueError(f"Merging NEQs with a common normal matrix but with varying sigma0 ratios over the auxiliary dimensions is not possible")
        # hasauxdim=True


    # var2_ratio=sigma_ratio_single**2
    # #check matrix compatibility
    # #check if N is in upper triangular or lower triangular form
    # if N1.attrs['xinv_state']!=N2.attrs['xinv_state']:
        # raise RuntimeError(f"Can't currently merge NEQs with different xinv_state:{N1.attrs['xinv_state']} vs {N2.attrs['xinv_state']}")
    
    # unkdim1=N1.dims[0]
    # unkdim2=N2.dims[0]
    
    # #find out the unique unknown parameters
    # # unique_unk_coord=pd.MultiIndex.from_tuples(np.unique(np.concatenate([N1[unkdim1].data,N2[unkdim2]])))
    # unique_unk_coord=xr.DataArray(np.unique(np.concatenate([N1[unkdim1].data,N2[unkdim2]])),dims=[unkdim1],name=unkdim1)
    # unique_unk_coord.attrs.update(xunk_coords_attrs(state=xinv_st.linked))

    # #find the auxiliary dimensions (ignore the unknown parameter dimension, and group_id/and seq)
    
    # xinvcoords=find_xinv_coords(dsneq,exclude=[xinv_tp.grp_id_co,xinv_tp.grp_seq_co])
    # group_id_co,group_seq_co=find_group_coords(dsneq)

    # #check if the group id and sequence coordinates are present in the first system
    # if group_id_co is not None and group_seq_co is not None:
        # #turn the unknown coordinate into a multiindex
        # grp_co=build_group_coord(unique_unk_coord,dim=unkdim1,group_id_name=group_id_co.name,group_seq_name=group_seq_co.name)
        # xinvcoords[unkdim1]=grp_co[unkdim1]
        # xinvcoords[group_id_co.name]=grp_co[group_id_co.name]
        # xinvcoords[group_seq_co.name]=grp_co[group_seq_co.name]
    # else:

        # #replace the unknow coordinate with the union version
        # xinvcoords[unkdim1]=unique_unk_coord
    # #add the proper attributes

    
    
    
    # xinvcoordsother=find_xinv_coords(dsneqother,exclude=[xinv_tp.grp_id_co,xinv_tp.grp_seq_co])
    # #possibly update with complementary coordinates from the second system
    # xinvcoords.update({ky:coord for ky,coord in xinvcoordsother.items() if ky not in xinvcoords.keys()})
    # #allocate space for the combined output normal equation system and use system one as the base template
    # dsneq_merged=xr.Dataset.xi.neqzeros(rhsdims=rhs1.dims,coords=xinvcoords)
    # #for some reason the multindex coordinate attributes do not get properly propagated
    # #so make sure they are added
    # for key,coord in xinvcoords.items():
        # if not dsneq_merged[key].attrs:
            # dsneq_merged[key].attrs.update(coord.attrs)

    # #rename the variables to the same name as in the first system
    # renamedict=dict(N=N1.name,rhs=rhs1.name,ltpl=ltpl1.name,sigma0=sigma01.name,nobs=nobs1.name,npara=npara1.name)
    # dsneq_merged=dsneq_merged.rename(renamedict)
    
    # #get links to the combined normal equation system
    # Nout,rhsout,x0out,ltplout,sigma0out,nobsout,nparaout=find_neq_components(dsneq_merged)
    # #Note: we need to be very careful when assigning values to the above views so that we insert data at the existing memory locations (!!), rather than replace the views alltogether
    # #add stuff from the first normal equation system
    
    # idx1=find_ilocs(dsneq_merged,unkdim1,N1[unkdim1].data)
    # Nout[idx1,idx1]=N1.data
    # rhsout[{unkdim1:idx1}]=rhs1.data
    
    # #note use [()] to replace the value in place
    # ltplout[()]=ltpl1
    # sigma0out[()]=sigma01
    # nobsout[()]=nobs1
    # nparaout[()]=np.int64(dsneq_merged.sizes[unkdim1])
    # #possibly add reduced parameters from the first system
    # nreduced=npara1-dsneq.sizes[unkdim1]
    # nparaout+=nreduced.astype(np.int64)


    
    # #add stuff from the second normal equation system
    # idx2=find_ilocs(dsneq_merged,unkdim2,N2[unkdim2].data)

    
    
    # #Note that we need to take care of possible different sigma0's
    # if var2_ratio != 1.0:
        # xinvlogger.info(f"Applying a rescaling of {var2_ratio} to the second normal equation system to compensate for the different apriori variances")
    
    # Nout[idx2,idx2]+=var2_ratio*N2.data
    
    # rhsout[{unkdim1:idx2}]+=var2_ratio*rhs2.data
    # ltplout+=var2_ratio*ltpl2
    # #note dsneq_merged['sigma0'] stays the same 
    # nobsout+=nobs2
    # #possibly add the amount of reduced parameters from the second system
    # nreduced=npara2-dsneqother.sizes[unkdim2]
    # nparaout+=nreduced.astype(np.int64)

    
    # return dsneq_merged                     
    


